require(ggplot2)
require(plotly)
require(data.table)
require(stringr)
require(dbplyr)

#setup
#data directory
dataBaseDir <- "W:/offlineFiles/EASy-concepts/reasoner/measures"
#how many initial in-experiment runs (out of 10) shall be skipped
skipRuns <- 3
#maximum number of full repetitions (entire experiments)
fullRepetitions <- 5
#numerical columns in data for which statistics shall be calculated
numCols <- c("RESPONSETIME_AUTO", "REASONER_EVALUATION_TIME", "REASONER_INSTANCE_CREATION_TIME", "REASONER_REASONING_TIME", "REASONER_TRANSLATION_TIME")
#names of files produced during measurement
filenames <- c('measurements-sse.tsv', 'measurements-rtvil.tsv', 'measurements-scenarios.tsv', 'measurements-scenarioVariants.tsv', 'measurements-scaleLog.tsv')
#tags in data to be skipped
skipTags <- c("CompoundInit")
#tag repacement (happens before skipping)
replaceTags <- data.frame(
  c("REASONING 1", "REASONING 2", "SCENARIO-INC",           "SCENARIO-INST",       "IREASONING"), 
  c("REASONING",   "REASONING",   "SCENARIO (incremental)", "SCENARIO (runtime)",  "REASONING (incremental)"))
names(replaceTags) <- c("orig", "subst")

# na-tolerant length function
my.lengthNA <- function (x, na.rm=FALSE) {
  if (na.rm) 
    sum(!is.na(x))
  else
    length(x)
}

#http://www.cookbook-r.com/Graphs/Plotting_means_and_error_bars_(ggplot2)/
# statistical summary over x, returns mean, median, standard deviation, standard error and confidence interval  
my.summary = function(x, conf.interval=.95) {
  n <- my.lengthNA(x, TRUE)
  res <- list(mean = mean(x), median = median(x), sd = sd(x), se=0, ci=0)
  res$se <- res$sd / sqrt(n)

  # Confidence interval multiplier for standard error
  # Calculate t-statistic for confidence interval: 
  # e.g., if conf.interval is .95, use .975 (above/below), and use df=N-1
  ciMult <- ifelse(n > 1, qt(conf.interval/2 + .5, n-1), 0) # 0 does not matter -> res$se anyway NA
  res$ci <- res$se * ciMult
  # see also https://www.cyclismo.org/tutorial/R/confidence.html (Section 9.2)
  return(res)
}

# shortens a given URI to the path segments including and after testdata (also including one path segment before testdata)
my.shortenURI = function(uri) {
  tmp <- str_match(uri, ".*/([^/]*/testdata/.*)")
  return (ifelse(is.na(tmp[1,2]), uri, tmp[1,2]))
}

#trims a given string x, i.e., removes trailing and leading whitespaces at
my.trim <- function (x) gsub("^\\s+|\\s+$", "", x)

# searches for tag in replaceTags and either returns the replacement for (trimmed) tag in replaceTags or tag (trimmed)
my.actualTag = function(tag) {
  tmp <- my.trim(tag)
  row <- filter(replaceTags, orig==tmp)
  if (nrow(row) == 1) {
    cells <- select(row, subst)
    res <- as.character(cells[[1]])
  } else {
    res <- tmp
  }
  return (res)
}

#reads a file produced by the MeasurementCollector of EASy-Producer
# in given folder
# using given filename (infix is constructed from run)
# actual run count for complete experiment repetition
# skipping warmup runs (default global skipRuns)
# statistical colums (default global numCols)
# tags to be skipped (default global skipTags)
#shortens URIs using my.shortenURI
#factors model.name, URI, tag, caller
#applies my.summary for statistical summaries on numCols
#projects to URI, tag, model.name, MODEL_COMPLEXITY and statistics of numCols
my.readFile = function(folder, filename, run=0, skip=skipRuns, statCols=numCols, skipT=skipTags) {
  if (run > 0)  
    tmp.folder <- paste(folder, run, sep='/')
  else
    tmp.folder <- folder
  #read the data, use header as colum names, turn them into real identifiers (check.names), avoid reading strings as factors 
  tmp.resTable <- as.data.table(read.table(file = paste(tmp.folder, filename, sep='/'), sep = '\t', header = TRUE, dec=",", check.names=TRUE, stringsAsFactors=FALSE)) 
  #preprocess, only runs later than given run count shall be in the result
  tmp.resTable <- tmp.resTable[runCount > skip]
  #preprocess, shorten URIs for interactive display
  #we must not have factors on URI, but do need factors on later operations as index
  tmp.resTable <- tmp.resTable[,URI:=factor(unlist(lapply(URI, my.shortenURI)))]
  tmp.resTable <- tmp.resTable[,model.name:=factor(model.name)]
  tmp.resTable <- tmp.resTable[,tag:=factor(unlist(lapply(tag, my.actualTag)))]
  tmp.resTable <- tmp.resTable[,caller:=factor(caller)]
  if (length(skipT) > 0) {
    tmp.resTable <- tmp.resTable[!(tag %in% skipTags)]
  }
  return (tmp.resTable)
}

#read all data for an experiment run.
# folder the folder to read from
# number of repetitions, 1 -> read files directly in folder, >1 read files from sub-folders named 1...n
#  default -> global fullRepetitions
#obtains the data, applies the (aggregated) statistics, returns one full table over all experiments
my.readData = function(folder, repetitions=fullRepetitions, skip=skipRuns, statCols=numCols, skipT=skipTags) {
  tmp.groupedTables <- list()
  tmp.count <- 0;
  for (f in 1:length(filenames)) {
    tmp.resTables <- list()
    if (repetitions == 1) {
        tmp.table <- my.readFile(folder, filenames[f], repetitions, skip, statCols, skipT) 
    } else {
      for (run in 1:fullRepetitions){
          #read the data
          tmp.resTable <- my.readFile(folder, filenames[f], run, skip, statCols, skipT) 
          #and "append" to list
          tmp.resTables[[run]] <- tmp.resTable
      }  
      #stack all tables for one-step analysis
      tmp.table<-rbindlist(tmp.resTables)
    }
    #group table by URI, tag and modelName, turn all cols in numCols/groups into table view .SD and apply summary function 
    #unpack result of summary so that names are taken over in colum names
    #see P. Tuma, do we teach the right statistics? that for some forms of statistics aggregation shall be ok
    tmp.table <- tmp.table[,as.list(unlist(lapply(.SD, my.summary))), list(URI, tag, model.name, MODEL_COMPLEXITY, REASONER_CONSTRAINT_COUNT), .SDcols=statCols]
    tmp.groupedTables[[f]] <- tmp.table
  }
  #stack all grouped tables
  data <- rbindlist(tmp.groupedTables)
  return (data);  
}

#return col in data, i.e., either access it by name if col is a string or return col directly
#(assuming that col is then something like data$col)
my.get = function(data, col) {
    if (is.character(col))
      res<-data[[col]]
    else
      res<-col
    return(res)
}

# emulate the ggplot2 colors for n given colors
#https://stackoverflow.com/questions/8197559/emulate-ggplot2-default-color-palette
gg_color_hue <- function(n) {
  hues = seq(15, 375, length = n + 1)
  hcl(h = hues, l = 65, c = 100)[1:n]
}

# draws a diagram for data using xCol (default MODEL_COMPLEXITY), y-axis and error-bar y values as given
# URI as label (for interactive use ggplotly(tooltip=c("label")) after plot(gg)) and tags as values.
# default settings can be changed through data$... or column names as strings
my.createErrorBarDiagram = function(data, yCol, yColName, yBarCol, diagramTitle="", xCol="MODEL_COMPLEXITY", xColName="model complexity", labelCol="URI", colorCol="tag", colorLegendName="", regression=TRUE, sizeFactor=5, legendPos="right") {
  size <- length(data)
  #ggplot seems to do levels(factor(.)) by its own
  legendFactors <- my.get(data, colorCol)
  #recalculate as otherwise legend is outdated
  legendLabels <- levels(factor(legendFactors))
  bar.width = size * sizeFactor # unclear, with seems to be correlated to the size, stackoverflow.com/questions/19420903/wideth-of-error-bars-in-ggplot2
  gg<-ggplot(data, aes(x=my.get(data, xCol),y=my.get(data, yCol),label=my.get(data, URI),color=legendFactors,shape=legendFactors)) + 
    geom_errorbar(aes(ymin=my.get(data, yCol)-my.get(data, yBarCol), ymax=my.get(data, yCol)+my.get(data, yBarCol)), width=bar.width) + 
    geom_point() +
    labs(x=xColName, y=yColName, title = diagramTitle, color=colorLegendName) +
    #override color and shape in initial aes to have only a single legend 
    #http://www.sthda.com/english/wiki/ggplot2-point-shapes
    #rep just to be sure if data changes
    scale_shape_manual(name="", labels=legendLabels, values=rep(c(0, 1, 2, 5, 6, 15, 17, 19))) +
    scale_color_manual(name="", labels=legendLabels, values=gg_color_hue(length(legendLabels))) +
    theme(legend.position=legendPos)
  if (regression) {
    gg <- gg + geom_smooth(method=lm)
  }
  return (gg)
}

#returns the last path segment of dir or an empty string
my.lastPathSegment = function(dir) {
  tmp <- str_match(dir, '.*/([^/]+)/?')
  return (ifelse(is.na(tmp[1,2]), "", tmp[1,2]))
}

# composes a diagram title from a type (name), an indicator (name) and the directory name
my.composeDiagramTitle = function(type, indicator, dir) {
  if (length(dir) > 0) {
    dir <- my.lastPathSegment(dir)
    #remove prefix
    dir <- str_remove(dir, "^benchmark-results-")
    #remove date postfix
    dir <- str_remove(dir, "-[:digit:]{8}$")
    dir <- paste(" ", dir, sep="")
  }
  return (paste(type, dir, " (", indicator, ")", sep=""))
}

# composes a diagram file name from a directory, a name and an indicator (name)
my.composeFileName = function(dir, name, indicator) {
  prefix <- my.lastPathSegment(dir)
  if (length(prefix) > 0) {
    prefix <- paste(prefix, "_", sep="")
  }
  name <- paste(prefix, name, "_", indicator, ".pdf", sep="")
  return (paste(dir, name, sep="/"))
}

my.printStats = function(data, name) {
  print(paste(name, min(data), mean(data), max(data), sep=" "))
}

my.printStatistics = function(data, dir, indicator) {
  print(paste("#tests", nrow(data), sep=" "))
  my.printStats(data$REASONER_REASONING_TIME.mean, "reasoning time")  
  my.printStats(data$REASONER_TRANSLATION_TIME.mean, "translation time")  
  my.printStats(data$REASONER_EVALUATION_TIME.mean, "evaluation time")  
  my.printStats(data$REASONER_INSTANCE_CREATION_TIME.mean, "instance creation time")  
  my.printStats(data$REASONER_CONSTRAINT_COUNT, "constraints count")  
}

# creates the diagrams for a data set to be written into dir using the given indicator (name)
my.createDiagrams = function(data, dir, indicator) {
  print(paste(" reasoning time diagram", indicator))
  gg <-
    my.createErrorBarDiagram(
    data, 
    data$REASONER_REASONING_TIME.mean, 
    "mean reasoning time", 
    data$REASONER_REASONING_TIME.ci, 
    diagramTitle=my.composeDiagramTitle("Reasoning Time", indicator, dir))
  ggsave(
    filename=my.composeFileName(dir, "reasoningTime", indicator), 
    plot=gg)
  
  print(paste(" translation time diagram", indicator))
  gg <- my.createErrorBarDiagram(
    data, 
    data$REASONER_TRANSLATION_TIME.mean, 
    "mean translation time", 
    data$REASONER_TRANSLATION_TIME.ci, 
    diagramTitle=my.composeDiagramTitle("Translation Time", indicator, dir))
  ggsave(
    filename=my.composeFileName(dir, "translationTime", indicator), 
    plot=gg)

  print(paste(" evaluation time diagram", indicator))
  gg <- my.createErrorBarDiagram(
    data, 
    data$REASONER_TRANSLATION_TIME.mean, 
    "mean evaluation time", 
    data$REASONER_TRANSLATION_TIME.ci, 
    diagramTitle=my.composeDiagramTitle("Evaluation Time", indicator, dir))
  ggsave(
    filename=my.composeFileName(dir, "evaluationTime", indicator), 
    plot=gg)
  
  my.printStatistics(data, dir, indicator)
}

# process the given directories, i.e., load data in folders starting with benchmark-results and
# draw/store diagrams
my.processDirs = function(dirs) {
  for (d in dirs) {
    if (!is.na(str_match(d, "benchmark-results.*"))) {
      print(paste("processing folder ", d))
      print(" loading data (1)")
      data <- my.readData(d, 1)
      my.createDiagrams(data, d, 1)
      print(" loading data (all)")
      data <- my.readData(d)
      my.createDiagrams(data, d, "all")
    }
  }
} 

# process all directores in dir using my.processDirs
my.processAll = function(dir) {
  sink(paste(dir, "log.txt", sep="/"))
  dirs <- list.dirs(path=dir, recursive=FALSE)
  my.processDirs(dirs)
  sink()
}

my.processAll(dataBaseDir)

# DEBUGGING STUFF

#dirs <- list.dirs(path="W:/offlineFiles/EASy-concepts/reasoner/measures", recursive=FALSE)

#data <- my.readFile('W:/offlineFiles/EASy-concepts/reasoner/measures/new', 'measurements-sse.tsv')
#data <- my.readData('W:/offlineFiles/EASy-concepts/reasoner/measures/benchmark-results-new-1.3.0-20190215', 1)
#gg <- my.createErrorBarDiagram(data, data$REASONER_REASONING_TIME.mean, "mean reasoning time", data$REASONER_REASONING_TIME.ci, diagramTitle="Reasoning time")
#ggplotly(tooltip=c("label"))


#gg<-ggplot(data, aes(x=MODEL_COMPLEXITY,y=REASONER_REASONING_TIME.mean,label=URI,color=tag)) + 
#  geom_errorbar(aes(ymin=REASONER_REASONING_TIME.mean-REASONER_REASONING_TIME.ci, ymax=REASONER_REASONING_TIME.mean+REASONER_REASONING_TIME.ci), size=1, width=50) + 
#  geom_smooth(method=lm, data=data) +
#  geom_point()
#ggplotly(tooltip=c("label"))